#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#if defined _WIN32 || defined __CYGWIN__
# include <windows.h>
#else
# include <sys/resource.h>
#endif
#if defined __ARC64__ || defined __hppa__ && defined __LP64__
# include <unistd.h>
#endif
#if defined __hppa__ && defined __LP64__
# include <fcntl.h>
#endif

#define MAX_SIZE	4194304

#pragma GCC optimize("O3")

long mrand48(void);
void srand48(long);

void downHeap(long long *a, int k, int n)
{
	long long tmp;
	int i;

	n--;
	for (tmp = a[k]; k <= n / 2; k = i) {
		i = 2 * k;
		if (i < n && a[i] < a[i+1])
			i++;
		if (tmp >= a[i])
			break;
		a[k] = a[i];
	}
	a[k] = tmp;
}

void heapSortC(long long *a, int n)
{
	long long tmp;
	int i;

	for (i = n / 2; i >= 0; i--)
		downHeap(a, i, n);
	for (i = n - 1; i > 0; i--) {
		tmp = a[i];
		a[i] = a[0];
		a[0] = tmp;
		downHeap(a, 0, i);
	}
}

#ifdef __amd64__
# define ATTRIB __attribute__((sysv_abi))
#else
# define ATTRIB
#endif

void ATTRIB heapSort(long long *, int);

int comp(const void *a, const void *b)
{
	return *(long long *)a - *(long long *)b;
}

long long useconds()
{
#ifdef __mmix__
	asm(
		"get	$0,rU\n\t"
		"slu	$0,$0,16\n\t"
		"sru	$0,$0,18\n\t"
		"pop	1,0\n"
	);
	return 0;
#elif defined __ia64__
	return clock();	/* CLOCKS_PER_SEC = 1000000 */
#elif defined _WIN32 || defined __CYGWIN__
	LARGE_INTEGER t, f;

	QueryPerformanceCounter(&t);
	QueryPerformanceFrequency(&f);
	return t.QuadPart * 1000000LL / f.QuadPart;
#elif defined __hppa__ && defined __LP64__
	int f;

	if ((f = open("/proc/stat", O_RDONLY)) > 0) {
		char buf[BUFSIZ];
		ssize_t n = read(f, buf, sizeof(buf));
		close(f);
		if (n > 0) {
			char *p = strchr(buf, ' ');
			if (p && p < buf + n)
				return atoll(p) * (1000000LL / 100);
		}
	}
	return time(NULL) * 1000000LL;
#else
	struct rusage ru;

	if (getrusage(RUSAGE_SELF, &ru)) {
		perror("getrusage()");
		exit(errno);
	}
	return (long long)(ru.ru_utime.tv_sec + ru.ru_stime.tv_sec) * 1000000LL
			+ ru.ru_utime.tv_usec + ru.ru_stime.tv_usec;
#endif
}

long long fill_in_and_start(long long *a, int n)
{
	size_t i;

	srand48(1);
	for (i = 0; i < n; i++)
		a[i] = mrand48() | (long long)mrand48() << 32;
	return useconds();
}

void print_time(const char *s, long long old)
{
	printf("%s: %.6f s\n", s, (double)(useconds() - old) / 1000000.);
}

int main()
{
	int i, size;
	long long start;
#ifdef __ARC64__
	static long long a[MAX_SIZE];
#elif defined __hppa__ && defined __LP64__
	long long *a = sbrk(MAX_SIZE * sizeof(long long));
#else
	long long *a = malloc(MAX_SIZE * sizeof(long long));
#endif
	setvbuf(stdout, NULL, _IONBF, 0);
	for (i = 0, size = MAX_SIZE / 1024; i < 11; i++, size *= 2) {
		printf("Array size %7d:", size);
		start = fill_in_and_start(a, size);
		qsort(a, size, sizeof(long long), comp);
		print_time("\tQuickSort", start);
		start = fill_in_and_start(a, size);
		heapSortC(a, size);
		print_time("\t\t\tHeapSortC", start);
		start = fill_in_and_start(a, size);
		heapSort(a, size);
		print_time("\t\t\t HeapSort", start);
	}
#ifdef __ARC64__
	_exit(0);
#elif defined __hppa__ && defined __LP64__
	brk(a);
	return 0;
#else
	free(a);
	return 0;
#endif
}
