#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

ssize_t getrandom(void *, size_t, unsigned);

#define SIZE	20

#if defined __LCC__ && __LCC__ < 124

typedef struct {
	unsigned long long lo;
	unsigned long long hi;
} __uint128_t;
typedef struct {
	__uint128_t lo;
	__uint128_t hi;
} __uint256_t;
__uint256_t umul128(__uint128_t, __uint128_t);
__uint128_t str2ulll(const char *, char **, int);
int srch128(__uint128_t, size_t, __uint128_t *);

int main()
{
	__uint128_t a[SIZE], t;
	char s[BUFSIZ];
	int i, j;

	if (getrandom(a, sizeof(a), 0) < 0)
		return -1;
	for (i = 0; i < SIZE; i++) {
		char *p = s + 40;
		t = a[i];
		int n = (int64_t)t.hi < 0;
		if (n) {
			if ((t.lo = -t.lo) != 0)
				t.hi++;
			t.hi = -t.hi;
			p--;
		}
		*p = '\0';
		do {
			unsigned long long x = t.lo;
			// H. Warren, Hacker's Delight, 10-14 Sample Magic Numbers
			__uint128_t m = {0xCCCCCCCCCCCCCCCD, 0xCCCCCCCCCCCCCCCC};
			__uint256_t y = umul128(t, m);
			t = y.hi;
			t.lo = t.hi << 61 | t.lo >> 3;
			t.hi >>= 3;
			*--p = x - 10 * t.lo + '0';
		} while (t.hi | t.lo);
		printf("%2d: ", i);
		for (j = 0; j < p - s; j++)
			putchar(' ');
		if (n)
			putchar('-');
		puts(p);
	}
	printf("Search for? ");
	fgets(s, sizeof(s), stdin);
        t = str2ulll(s, NULL, 10);
	if ((i = srch128(t, SIZE, a)) == -1)
		printf("Not found!\n");
	else
		printf("Found at index %d\n", i);
	return 0;
}

#else

#ifdef __ARC64__
# include <unistd.h>
#endif

#ifdef __amd64__
# define ATTRIB __attribute__((sysv_abi))
#else
# define ATTRIB
#endif

__uint128_t str2ulll(const char *, char **, int);
int ATTRIB srch128(__int128_t, size_t, __int128_t *);

int main()
{
	__int128_t a[SIZE], t;
	int i, j;
	char s[BUFSIZ], *p;

	setvbuf(stdin, NULL, _IONBF, 0);
	setvbuf(stdout, NULL, _IONBF, 0);
#if defined __APPLE__ || defined __OpenBSD__
        arc4random_buf(a, sizeof(a));
#else
	getrandom(a, sizeof(a), 0);
#endif
	for (i = 0; i < SIZE; i++) {
		__uint128_t n = a[i] < 0 ? -a[i] : a[i];
		p = s + 39;
		*p = '\0';
		do
			*--p = n % 10 + '0';
		while (n /= 10);
		printf("%2d: ", i);
		for (j = 0; j < p - s; j++)
			putchar(' ');
		putchar(a[i] < 0 ? '-' : ' ');
		puts(p);
	}
	printf("Search for? ");
	fgets(s, sizeof(s), stdin);
        t = str2ulll(s, &p, 10);
	if ((i = srch128(t, SIZE, a)) == -1)
		printf("Not found!\n");
	else
		printf("Found at index %d\n", i);
#ifdef __ARC64__
	_exit(0);
#else
	return 0;
#endif
}

#endif
